/* Musical Blob
 * University of Washington Computer Science and Engineering
 * Sound Capstone (CSE 481i)
 * Project Members: Nathan Armstrong, Daniel Chesney, Andrew Foster
 *
 * CliqueBlobSystem
 * For documentation, see CliqueBlobSystem.hpp
 */

#include <algorithm>
#include <cmath>
#include "CliqueBlobSystem.hpp"

using std::min;
using std::max;
using std::vector;

//Determines the scale by which spring motion is amplified
#define DRAG_CONSTANT 0.00005//0.005//0.1//0.005//0.05
//Determines the scale by which motion is dampened
#define FRICTION_SCALE 3//0.9//0.6 //0.9 //0.4
//Determines the rest position for springs
#define SPRING_REST_DIST 100
//Determines the margin around the rest position at which the spring will
//remain relaxed
#define SPRING_REST_MARGIN 40
//Determines the distance of maximum pull for the spring
#define SPRING_PULL 150
//Determines the scale of the rebound force from wall collision
//(No longer used)
#define RECOIL_MAG 100
//Determines the scale of the rebound force from wall collision
#define COLL_DAMP 0.6

CliqueBlobSystem::~CliqueBlobSystem(){
	for(vector<Particle*>::iterator i = parts.begin();
		i != parts.end(); i++)
		delete (*i);
}

void CliqueBlobSystem::addParticle(Particle& p){
	parts.push_back(&p);
}

vector<Particle*> CliqueBlobSystem::getParticles() const{
	return vector<Particle*>(parts);
}

void CliqueBlobSystem::setRoomBounds(double nxMax, double nyMax,
                                     double nzMax){
	xMax = nxMax;
	yMax = nyMax;
	zMax = nzMax;
}

void CliqueBlobSystem::updateForces(){
	vector<Particle*>::const_iterator iteri, iterj;

	//YUCK N^2
	for(iteri = parts.begin(); iteri != parts.end(); iteri++){
		applySprings(*iteri);
		applyFriction(*iteri);
		applyJiggle(*iteri);
		applyRoomCollision(*iteri);
	}
}

void CliqueBlobSystem::applyForces(){
	vector<Particle*>::const_iterator iteri;
	for(iteri = parts.begin(); iteri != parts.end(); iteri++){
		(*iteri)->updateParticle(0.005);
		(*iteri)->clearForceAccum();
	}
}

void CliqueBlobSystem::applySprings(Particle* p){
	vector<Particle*>::const_iterator iteri;
	for(iteri = parts.begin(); iteri != parts.end(); iteri++){
		Particle* o = *iteri;
		if(p != o){
			TripleD offset = (o->getCurrPos()) - (p->getCurrPos());
			double d = offset.magnitude() - SPRING_REST_DIST;
			if(d < -SPRING_REST_MARGIN) d += SPRING_REST_MARGIN;
			else if(d > SPRING_REST_MARGIN) d -= SPRING_REST_MARGIN;
			else return;
			if((d < -SPRING_REST_MARGIN || d > SPRING_REST_MARGIN)
					&& d < (2*SPRING_PULL) && d != 0){
				static double quadScale = 0.5 / SPRING_PULL;
				d *= quadScale;
				double scale = 4 * d * (1 - d);//(1 - (SPRING_REST_DIST / d)) * DRAG_CONSTANT;
				p->addForce(offset.scale(scale));
			}
		}
	}
}

void CliqueBlobSystem::applyFriction(Particle* p){
	TripleD velocity = (p->getCurrPos()) - (p->getPrevPos());
	p->addForce(velocity.scale(-FRICTION_SCALE));
}

void CliqueBlobSystem::applyJiggle(Particle* p){
	p->addForce(TripleD((rand() % 21) - 10,
		                (rand() % 21) - 10,
		                (rand() % 21) - 10));
}

TripleD clampTriple(TripleD& t, double xMax, double yMax, double zMax){
	return TripleD(min(max(t.getElemA(), 0.0), xMax),
		           min(max(t.getElemB(), 0.0), yMax),
				   min(max(t.getElemC(), 0.0), zMax));
}

void CliqueBlobSystem::applyRoomCollision(Particle* p){
	TripleD currPos = p->getCurrPos(), prevPos = p->getPrevPos();
	if(currPos.getElemA() < 0){
		currPos = TripleD(-currPos.getElemA(),
			              currPos.getElemB(),
						  currPos.getElemC());
		prevPos = TripleD(-prevPos.getElemA(),
			              prevPos.getElemB(),
						  prevPos.getElemC());
		prevPos = (prevPos - currPos).scale(COLL_DAMP) + currPos;
	}
	if(currPos.getElemA() > xMax){
		currPos = TripleD((2 * xMax) - currPos.getElemA(),
			              currPos.getElemB(),
						  currPos.getElemC());
		prevPos = TripleD((2 * xMax) - prevPos.getElemA(),
			              prevPos.getElemB(),
						  prevPos.getElemC());
		prevPos = (prevPos - currPos).scale(COLL_DAMP) + currPos;
	}
	if(currPos.getElemB() < 0){
		currPos = TripleD(currPos.getElemA(),
			              -currPos.getElemB(),
						  currPos.getElemC());
		prevPos = TripleD(prevPos.getElemA(),
			              -prevPos.getElemB(),
						  prevPos.getElemC());
		prevPos = (prevPos - currPos).scale(COLL_DAMP) + currPos;
	}
	if(currPos.getElemB() > yMax){
		currPos = TripleD(currPos.getElemA(),
			              (2 * yMax) - currPos.getElemB(),
						  currPos.getElemC());
		prevPos = TripleD(prevPos.getElemA(),
			              (2 * yMax) - prevPos.getElemB(),
						  prevPos.getElemC());
		prevPos = (prevPos - currPos).scale(COLL_DAMP) + currPos;
	}
	if(currPos.getElemC() < 0){
		currPos = TripleD(currPos.getElemA(),
			              currPos.getElemB(),
						  -currPos.getElemC());
		prevPos = TripleD(prevPos.getElemA(),
			              prevPos.getElemB(),
						  -prevPos.getElemC());
		prevPos = (prevPos - currPos).scale(COLL_DAMP) + currPos;
	}
	if(currPos.getElemC() > zMax){
		currPos = TripleD(currPos.getElemA(),
			              currPos.getElemB(),
						  (2 * zMax) - currPos.getElemC());
		prevPos = TripleD(prevPos.getElemA(),
			              prevPos.getElemB(),
						  (2 * zMax) - prevPos.getElemC());
		prevPos = (prevPos - currPos).scale(COLL_DAMP) + currPos;
	}
	p->setCurrPos(clampTriple(currPos, xMax, yMax, zMax));
	p->setPrevPos(prevPos);
}
