/* Musical Blob
 * University of Washington Computer Science and Engineering
 * Sound Capstone (CSE 481i)
 * Project Members: Nathan Armstrong, Daniel Chesney, Andrew Foster
 *
 * OrbitalBlobSystem
 * For documentation, see OrbitalBlobSystem.hpp
 */

#include <algorithm>
#include "OrbitalBlobSystem.hpp"

//Helps determine how attracted particles are to each other
#define GRAVITY_PULL 2
//Determines the scale by which spring motion is amplified
#define DRAG_CONSTANT 0.1//0.005//0.05
//Determines the scale by which motion is dampened
#define FRICTION_SCALE 0.4 //0.9 //0.4
//Determines the rest position for springs
#define SPRING_REST_DIST 200
//Determines the scale of the rebound force from wall collision
//(No longer used)
#define RECOIL_MAG 1
//Determines the scale of the rebound force from wall collision
#define COLL_DAMP 0.6

using std::min;
using std::max;

OrbitalBlobSystem::~OrbitalBlobSystem(){
	for(std::vector<Particle*>::iterator i = parts.begin();
		i != parts.end(); i++)
		delete (*i);
}

void OrbitalBlobSystem::addParticle(Particle& p){
	parts.push_back(&p);
}

std::vector<Particle*> OrbitalBlobSystem::getParticles() const{
	return std::vector<Particle*>(parts);
}

void OrbitalBlobSystem::setRoomBounds(double nxMax, double nyMax, double nzMax){
	xMax = nxMax;
	yMax = nyMax;
	zMax = nzMax;
}

void OrbitalBlobSystem::updateForces(){
	std::vector<Particle*>::const_iterator iteri, iterj;

	applyCentripetal();
	applySprings();
	for(iteri = parts.begin(); iteri != parts.end(); iteri++){
		applyFriction(*iteri);
		applyJiggle(*iteri);
		applyRoomCollision(*iteri);
	}
}

void OrbitalBlobSystem::applyForces(){
	std::vector<Particle*>::const_iterator iteri;
	for(iteri = parts.begin(); iteri != parts.end(); iteri++){
		(*iteri)->updateParticle(0.01);
		(*iteri)->clearForceAccum();
	}
}

void OrbitalBlobSystem::applyCentripetal(){
	static double centroidDiv = 1.0/parts.size();
	std::vector<Particle*>::const_iterator iteri;
	TripleD centroid;

	for(iteri = parts.begin(); iteri != parts.end(); iteri++)
		centroid += (*iteri)->getCurrPos();

	centroid = centroid.scale(centroidDiv);
	for(iteri = parts.begin(); iteri != parts.end(); iteri++){
		Particle* p = *iteri;
		TripleD diff = centroid - p->getCurrPos();
		double diffMag = diff.magnitude();
		diffMag *= diffMag;
		p->addForce(diff.scale(GRAVITY_PULL / diffMag));
	}
}

void OrbitalBlobSystem::applySprings(){
	if(parts.size() <= 1) return;
	Particle* first = parts[0];
	Particle* prev = first;

	std::vector<Particle*>::const_iterator iteri;
	for(iteri = parts.begin() + 1; iteri != parts.end(); iteri++){
		Particle* o = *iteri;
		TripleD offset = (o->getCurrPos()) - (prev->getCurrPos());
		double d = offset.magnitude();
		if(d){
			double scale = (1 - (SPRING_REST_DIST / d)) * DRAG_CONSTANT;
			prev->addForce(offset.scale(scale));
			o->addForce(offset.scale(-scale));
		}
		prev = o;
	}

	if(first != prev){
		TripleD offset = (first->getCurrPos()) - (prev->getCurrPos());
		double d = offset.magnitude();
		if(d){
			double scale = (1 - (SPRING_REST_DIST / d)) * DRAG_CONSTANT;
			prev->addForce(offset.scale(scale));
			first->addForce(offset.scale(-scale));
		}
	}
}

void OrbitalBlobSystem::applyFriction(Particle* p){
	TripleD velocity = (p->getCurrPos()) - (p->getPrevPos());
	p->addForce(velocity.scale(-FRICTION_SCALE));
}

void OrbitalBlobSystem::applyJiggle(Particle* p){
	p->addForce(TripleD((rand() % 21) - 10,
			            (rand() % 21) - 10,
						(rand() % 21) - 10));
}

extern TripleD clampTriple(TripleD& t, double xMax, double yMax, double zMax);

void OrbitalBlobSystem::applyRoomCollision(Particle* p){
	TripleD currPos = p->getCurrPos(), prevPos = p->getPrevPos();
	if(currPos.getElemA() < 0){
		currPos = TripleD(-currPos.getElemA(),
			              currPos.getElemB(),
						  currPos.getElemC());
		prevPos = TripleD(-prevPos.getElemA(),
			              prevPos.getElemB(),
						  prevPos.getElemC());
		prevPos = (prevPos - currPos).scale(COLL_DAMP) + currPos;
	}
	if(currPos.getElemA() > xMax){
		currPos = TripleD((2 * xMax) - currPos.getElemA(),
			              currPos.getElemB(),
						  currPos.getElemC());
		prevPos = TripleD((2 * xMax) - prevPos.getElemA(),
			              prevPos.getElemB(),
						  prevPos.getElemC());
		prevPos = (prevPos - currPos).scale(COLL_DAMP) + currPos;
	}
	if(currPos.getElemB() < 0){
		currPos = TripleD(currPos.getElemA(),
			              -currPos.getElemB(),
						  currPos.getElemC());
		prevPos = TripleD(prevPos.getElemA(),
			              -prevPos.getElemB(),
						  prevPos.getElemC());
		prevPos = (prevPos - currPos).scale(COLL_DAMP) + currPos;
	}
	if(currPos.getElemB() > yMax){
		currPos = TripleD(currPos.getElemA(),
			              (2 * yMax) - currPos.getElemB(),
						  currPos.getElemC());
		prevPos = TripleD(prevPos.getElemA(),
			              (2 * yMax) - prevPos.getElemB(),
						  prevPos.getElemC());
		prevPos = (prevPos - currPos).scale(COLL_DAMP) + currPos;
	}
	if(currPos.getElemC() < 0){
		currPos = TripleD(currPos.getElemA(),
			              currPos.getElemB(),
						  -currPos.getElemC());
		prevPos = TripleD(prevPos.getElemA(),
			              prevPos.getElemB(),
						  -prevPos.getElemC());
		prevPos = (prevPos - currPos).scale(COLL_DAMP) + currPos;
	}
	if(currPos.getElemC() > zMax){
		currPos = TripleD(currPos.getElemA(),
			              currPos.getElemB(),
						  (2 * zMax) - currPos.getElemC());
		prevPos = TripleD(prevPos.getElemA(),
			              prevPos.getElemB(),
						  (2 * zMax) - prevPos.getElemC());
		prevPos = (prevPos - currPos).scale(COLL_DAMP) + currPos;
	}
	p->setCurrPos(clampTriple(currPos, xMax, yMax, zMax));
	p->setPrevPos(prevPos);
}
