/* Musical Blob
 * University of Washington Computer Science and Engineering
 * Sound Capstone (CSE 481i)
 * Project Members: Nathan Armstrong, Daniel Chesney, Andrew Foster
 *
 * OscMessageSender
 * For documentation, see OscMessageSender.hpp
 */

#include <limits>
#include <algorithm>
#include <sstream>
#include <string.h>
#include "OscMessageSender.hpp"

#define BUFFER_SIZE 1024

using std::min;
using std::max;
using std::vector;

const string INIT_TAG("/init");
const string CLEAR_TAG("/clear");
const string NODE_MSG_TAG("/nodeMsg");
const string GLOBAL_TAG("/global");

//Clamps the given value to the range [0,1]
double unitClamp(double d){
	return min(max(d, 0.0), 1.0);
}

//Clamps all values in the given triple to the range [0,1]
TripleD unitClamp(TripleD t){
	return TripleD(unitClamp(t.getElemA()), unitClamp(t.getElemB()),
		           unitClamp(t.getElemC()));
}

//Returns a string representation of the given integer
string intToString(int i){
	std::ostringstream stream;
	stream << i;
	string s = stream.str();
	return s;
}

OscMessageSender::OscMessageSender(const char* ip, unsigned int port):
		sock(IpEndpointName(ip, port)), buffer(new char[BUFFER_SIZE]){
	stream = new osc::OutboundPacketStream(buffer, BUFFER_SIZE);
}

OscMessageSender::~OscMessageSender(){
	delete[] buffer;
	buffer = NULL;
	delete stream;
	stream = NULL;
}

void OscMessageSender::sendInit(int numParts, SoundSynth synth){
	sendClear();
	stream->Clear();
	startMessage(INIT_TAG);
	(*stream) << numParts << (int)synth;
	endMessage();
	sock.Send(stream->Data(), stream->Size());
}

void OscMessageSender::sendClear(){
	stream->Clear();
	startMessage(CLEAR_TAG);
	endMessage();
	sock.Send(stream->Data(), stream->Size());
}

void OscMessageSender::sendParticleBatch(std::vector<Particle*> parts,
		double xScale, double yScale, double zScale){
	vector<Particle*>::const_iterator i;
	TripleD centroid, centImp;

	double minX, minY, minZ;
	minZ = minY = minX = std::numeric_limits<double>::max();
	double maxX, maxY, maxZ;
	maxX = maxY = maxZ = 0;

	//Calculate the various components to send: centroid
	for(i = parts.begin(); i != parts.end(); i++){
		Particle* p = *i;
		TripleD curr = p->getCurrPos();

		centroid += curr;
		centImp += p->getForceAccum();
		minX = min(minX, curr.getElemA());
		minY = min(minY, curr.getElemB());
		minZ = min(minZ, curr.getElemC());
		maxX = max(maxX, curr.getElemA());
		maxY = max(maxY, curr.getElemB());
		maxZ = max(maxZ, curr.getElemC());
	}

	double invNumParts = 1.0/(parts.size());
	centroid = centroid.scale(invNumParts);
	centroid = unitClamp(centroid.scale(xScale, yScale, zScale));

	centImp = unitClamp(centImp.scale(0.0005) + TripleD(0.5, 0.5, 0.5));

	TripleD minCorner(minX * xScale, minY * yScale, minZ * zScale);
	TripleD maxCorner(maxX * xScale, maxY * yScale, maxZ * zScale);
	minCorner = unitClamp(minCorner);
	maxCorner = unitClamp(maxCorner);

	stream->Clear();

	unsigned int partSize = parts.size();
	for(unsigned int i = 0; i < partSize; i++){
		startMessage(NODE_MSG_TAG + intToString(i));
		TripleD pos = parts[i]->getCurrPos();
		TripleD imp = parts[i]->getForceAccum();

		sendTripleD(unitClamp(pos.scale(xScale, yScale, zScale)));
		sendTripleD(unitClamp(imp.scale(0.0005) + TripleD(0.5, 0.5, 0.5)));

		unsigned int triSuccIndex = (i/3) + ((i+1)%3);
		if(triSuccIndex < parts.size()){
			TripleD vel = pos - parts[i]->getPrevPos();
			TripleD toSucc = parts[triSuccIndex]->getCurrPos() - pos;
			if(vel.magnitude() == 0 || toSucc.magnitude() == 0)
				(*stream) << 0.0;
			else
				(*stream) << (vel.normalize() * toSucc.normalize());
		}

		endMessage();
	}

	startMessage(GLOBAL_TAG);
	sendTripleD(centroid);
	sendTripleD(centImp);
	sendTripleD(minCorner);
	sendTripleD(maxCorner);
	endMessage();
}

void OscMessageSender::startMessage(string s){
	char* temp = new char[s.size() + 1];
	strncpy(temp, s.c_str(), s.size());
	temp[s.size()] = '\0';
	(*stream) << osc::BeginMessage(temp);
}

void OscMessageSender::sendTripleD(const TripleD& t){
	(*stream) << t.getElemA() << t.getElemB() << t.getElemC();
}

void OscMessageSender::endMessage(){
	(*stream) << osc::EndMessage;
	sock.Send(stream->Data(), stream->Size());
	stream->Clear();
}
